import numpy as np
from  scipy.spatial.distance import pdist
import functools
import warnings

from .variogram import *
from .fvariogram import *

class Revarie:
    def __init__(self, x, mu, sill, model):
        """
        Class to generate random fields based on a variogram given as a
        function in the 'model' parameter, mean and variance for a number of
        points given in the x parameter.

        Parameters
        ----------
        x : numpy.ndarray
            Array of shape (m,n) where n is the number of points in an
            m-dimensional domain. Each row is a point. This does not need to
            be the same points used to calculate the original variogram.
        mu : float
            Spatially-independent mean of field values
        sill : float
            Spatially-independent variance of field values
        model : function
            Callable with takes numpy array of lag distances as argument and
            returns numpy array of variogram values. Should only take a single
            parameter.
        """
        self.x = x
        self.mu = mu
        self.sill = sill
        self.model = model
        self.s = x.shape[0]

        self.check_init()

        self.calc_cov(x, model)

    def calc_cov(self, x, model):
        """
        Creates the covariance matrix for the points with the variogram model
        """
        if x.ndim < 2:
            x = x.reshape(x.size, 1)

        h_cov = np.zeros((self.s, self.s))
        h_cov[np.diag_indices(self.s)] = self.sill
        mask_indices = np.mask_indices(self.s, np.triu, k=1)

        lags = pdist(x)

        h_cov[mask_indices] = self.sill - self.model(lags)

        self.cov = np.maximum(h_cov,h_cov.T)

    def genf(self):
        """
        Generates random field values using covariance matrix calculated
        previously.

        Returns
        -------
        fvs : numpy array
            Array of field values of length x.shape[0] that correspond to each
            point in x
        """

        fvs = np.random.multivariate_normal(np.ones(self.s)*self.mu, self.cov)
        return fvs

    def check_init(self):
        if not callable(self.model):
            raise Exception("Model initialization parameter must be function"
                    "which takes numpy array of lags as arg and returns corre"
                    "sponding variogram values")
        try:
            self.model(np.zeros(4))
        except:
            raise Exception("Lags will be passed as numpy array to callable d"
                    "efined in model input parameter. Should return numpy arr"
                    "ay as well")




import numpy as np
import warnings
from functools import partial, wraps
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from .models import * #mtags variable comes from here

def _fvariogram(f):
    @wraps(f)
    def wrapper(source, method, options, *args, **kwargs):
        if source == "func":
            if method == "ufunc":
                if not isinstance(options, list) and callable(options):
                    warnings.warn("User-defined function defined in options"
                        " parameter being reformatted as a single element "
                        "list, it is recommended that the user simply enclose"
                        "user-function in list before passing as argument to "
                        "options parameter.")
                    options = [options]
                if len(options) == 1:
                    try:
                        res = options[0](np.zeros(4))
                    except:
                        raise Exception("Lags will be passed as numpy array"
                        "user-defined function. User-defined functions also m"
                        "ust only take one argument")

            elif method in mtags.keys():
                try:
                    mtags[method](*([0] + options))
                except:
                    raise Exception("options parameter not of proper length "
                        "to fill *args of selected built-in model")
            else:
                n = ",".join(["'{g}'".format(g=h) for h in mtags.keys()])
                raise Exception("Method argument for 'func' source sould be "
                    "either 'ufunc' for user specified function or one of the"
                    " built-in models: " + n)

        elif source == "data":
            if method == "poly":
                pass
            elif method == "interp":
                pass
            elif method == "bmodel":
                if options[2] not in mtags.keys():
                    raise Exception("{g} not valid parameter for built-in mod"
                        "el".format(g = options[2]))
            elif method == "umodel":
                pass #this is where we should run validity tests AFTER fitting
        else:
            raise Exception("{s} is not a valid value for source parameter,"
                " only 'func' and 'data' are acceptable".format(s=source))

        return f(source, method, options, *args, **kwargs)
    return wrapper


@_fvariogram
def fvariogram(source, method, options, *args, **kwargs):
    """
    *Function intended to make generating python functions to describe
    variograms centralized. This function can be used to access the built-in
    odels as well as fit around user-defined models. Functionality includes
    fitting variograms to built-in and user-defined models, interpolation and
    polynomial fitting. Put simply, can take a wide range of parameters to
    describe how a function is made that gives variogram as a function of
    distance. Then, returns that function.

    Parameters
    ----------
    source : str
        Descriptor of the general approach used for the variogram model. Can
        be one of:
            * "func" : it is desired to return either a user-defined
                function, or a built-in function with specified parameters.
                This is not the option to select if any sort of fitting is
                desired. See below.
            * "data" : a model will be fit to data that is provided in the
                "options" parameter according to a method given in the
                "method" parameter.
    method : str
        Secondary descriptor which depends on the route selected in the source
        parameter. If "func" was selected, can be one of:
            * "ufunc" : user-specified python function object that takes
                numpy array of lag values and returns variogram values.Should
                only take a single object.
            * built-in model : pass a string of any tag associated with a
                built-in model to use that model with parameters specified
                later in the "options" parameter.

        If "data" is selected, can be one of:
            * "poly" : uses numpy polynomial fit tool to fit a polynomial to
                data provided later in the "options" parameter
            * "interp" : uses the scipy interpolate tools to interpolate data
                provided later in the "options" parameter
            * "bmodel" : fits the data provided in the "options" parameter to
                one of the built-in models using nonlinear least squares
            * "umodel" : fits the data provided in the "options" parameter to
                a user-defined function with an arbitrary number of arguments
                using nonlinear least squares. If values of the arguments are
                known, consider using the source = "func" route because this
                method will fit function parameters to provided data.
    options : list
        List of data needed for the options selected in the previous
        parameters. Descriptions of the content of these lists are given
        below:
            * "func" -> "ufunc" : [<userfunction>, *args, **kwargs]
                <userfunction> user-defined function to be called with
                an array of lags as input that returns an array of variogram
                values. Include extra args to function in *args and **kwargs.
            * "func" -> built-in model : [*args, **kwargs]
                list of parameters to be passed directly to the selected
                built-in model
            * "data" -> "poly" : [<h>, <v>, <opt>, <order>]
                <h> array of lag values to fit polynomial to
                <v> array of variogram values to fit polynomial to
                <order> order of polynomial for fitting as int
            * "data" -> "interp" : [<h>, <v>, <kind>]
                <h> array of lag values to use for interpolation
                <v> array of variogram values to use for interpolation
                <kind> str to specify how interpolation is performed, options
                include: "linear", "nearest", "zero", "slinear", "quadratic",
                "cubic", "previous" or "next". See scipy.interpolate.interp1d
                for more details
            * "data" -> "bmodel" : [<h>, <v>, <model tag>, <opt>]
                <h> array of lag values to fit model to
                <v> array of variogram values to fit model to
                <model tag> string specifying which of the built in models to
                fit
                <opt> Optional if True, also returns optimized parameters
            * "data" -> "umodel" : [<h>, <v>, <user func>, <opt>]
                <h> array of lag values to fit model to
                <v> array of variogram values to fit model to
                <user func> function object of to use for fitting, first arg
                must be lag values.
                <opt> Optional, if True returns optimized parameters
        *args/**kwargs
            Extra parameters to be passed into external functions for fitting
            or interpolation
                * "data" -> "poly"
                    Passed into numpy.polyfit function
                * "data" -> "interp"
                    Passed into the scipy.interpolate function
                * "data" -> "bmodel"
                    Passed into scipy.curve_fit
                * "data" -> "umodel"
                    Passed into scipy.curve_fit

            Returns
            -------
            fvariogram : function
                Callable function to be used in later variogram calculations.
                All returned functions will take one argument, lag distance h
                as an array. The return type of all returned functions will be
                an array of returned variogram values at each of those lag
                distances.

    """
    if source == "func":
        if method == "ufunc":
            f = options[0]
            if len(options) == 1:
                return f
            else:
                _f = lambda *a : f(*a[::-1])
                return partial(_f, *options[1::-1])
        else:
            f = mtags[method]
            _f = lambda *a : f(*a[::-1])
            return partial(_f, *options[::-1])

    elif source == "data":
        h = options[0]
        v = options[1]

        if len(options) < 4:
            options.append(False)#default to no optimized parameters return

        if method == "interp":
            return interp(h, v, options[2], *args, **kwargs)
        elif method == "poly":
            return polyfit(h, v, options[2], options[3], *args, **kwargs)
        elif method == "bmodel":
            return bmodel_fit(h, v, options[2], options[3], *args, **kwargs)
        elif method == "umodel":
            return umodel_fit(h, v, options[2], options[3], *args, **kwargs)

def interp(h, v, kind, *args, **kwargs):
    """
    See "data" -> "interp" above
    """
    return interp1d(h, v, kind, *args, **kwargs)

def polyfit(h, v, order, opt, *args, **kwargs):
    """
    See "data" -> "poly" above
    """
    P = np.polyfit(h,v,order, *args, **kwargs)

    def f(h):
        return np.polyval(P, h)

    if opt:
        return f, P
    else:
        return f

def bmodel_fit(h, v, model, opt, *args, **kwargs):
    """
    See "data" -> "bmodel" above
    """
    m = mtags[model]
    return umodel_fit(h, v, m, opt, *args, **kwargs)


def umodel_fit(h, v, f, opt, *args, **kwargs):
    """
    See "data" -> "umodel" above
    """
    opts, _ = curve_fit(f, h, v, *args, **kwargs)
    _f = lambda *a : f(*a[::-1])
    if opt:
        return partial(_f, *opts[::-1]), opts
    else:
        return partial(_f, *opts[::-1])



import numpy as np
from scipy.spatial.distance import pdist
from numpy_indexed import group_by
import functools
import warnings

from .fvariogram import fvariogram

class Variogram:
    """

    Calculates lag and squared difference values. Various operations can be
    performed with these quantities.

    """
    def __init__(self, x, f):
        """
        Create variogram and calculate lags and squared differences

        Parameters
        ----------
        x : numpy.ndarray
            Array of shape (m,n) where n is number of points in an
            m-dimensional domain
        f : numpy.ndarray
            Array of field values observed at each of the n points.
        """
        self.x = x
        self.f = f

        self.check_init()
        self.cond_init()

        self.s = f.size

        self.lags = self.calc_lags()
        self.diffs = self.calc_diffs()

        self.range = (np.min(self.lags), np.max(self.lags))

        self.reduced = False

    def cloud(self):
        """
        *Convenient way to get calculated lags and squared differences from
        a Variogram instance. Named for variogram cloud plot.

        Returns
        -------
        lags : numpy.ndarray
            Distance between all combinations of points fed into variogram.
        diffs : numpy.ndarray
            Squared difference between all combinations of field values fed
            into variogram
        """
        return self.lags, self.diffs

    def _c_matheron(f):
        @functools.wraps(f)
        def wrapper(self, bin_type = "auto", bins = 10, var = False):
            if bin_type == "auto":
                if not int(bins) == bins:
                    raise Exception("Number of bins not correctly castable to"
                            " int")
            elif bin_type == "bound":
                if not all(bins[i] <= bins[i+1] for i in range(len(bins)-1)):
                    raise Exception("Bin boundaries must be arranged in"
                                    " ascending order")
            elif bin_type == "lin":
                if bins[0] > bins[1]:
                    raise Exception("Check format of 'bins' parameter, minima"
                        " should be first element.")
                if int(bins[2]) != bins[2]:
                    raise Exception("Number of bins not correctly castable to"
                            " int")
            else:
                raise Exception("Not known bin type, either auto, bound or li"
                    "n.")

            return f(self, bin_type, bins, var)
        return wrapper

    @_c_matheron
    def matheron(self, bin_type = "auto", bins = 10, var = False):
        #should add binning with constant number of values
        """
        *Calculate Matheron variogram for points and field values previousely
        fed into variogram. A few options for specifying binning exist.

        Parameters
        ----------
        bin_type : str
            Descriptor of the format of data passed into the bin parameter.
            Can be one of:
                * "auto" : select bounds to be (0, self.range[1]/2), bin
                    centers will be calculated accordingly based on user given
                    number of bins.
                * "lin" : bin boundaries will be linearly spaced based on a
                    user given minima, maxima and number of bins. Bin centers
                    will not fall on given maxima and minima.
                * "bound" : bounds will be completely given by the user as a
                    numpy array
        bins : int, list, array-like
            Description of how binning will be performed in Matheron
            variogram, specific formats given below for each bin type.
                * "auto" : int giving number of bins to use
                * "lin" : list containing three entries. First is minima of
                    bin boundaries, second is maxima of bin bounds and third
                    is the number of bins to use as int
                * "bound" : array-like object specifying boundaries of bins.
                    Number of bins is length of this array - 1.
        var : bool
            Set True for bin-wise variance to be calculated and returned

        Returns
        -------
        centers : numpy.array
            Bin centers used for variogram
        n_bins : numpy array
            Number of point relations used to calculate each semivariance
        v : numpy array
            Estimated semivariance values at lags corresponding to bin centers
        v_var (optional) : numpy array
            Variance associated with squared difference values within a bin
        """
        bins = self.set_bins(bin_type, bins)
        centers = bins[:-1] + np.diff(bins,1)/2

        b_ind = np.digitize(self.lags, bins)
        n_bins = np.bincount(b_ind-1)[:-1]

        gp = group_by(b_ind[np.where(b_ind != bins.size)])
        _, v = gp.mean(self.diffs[np.where(b_ind != bins.size)])#account for lags bigger than bins

        if var:
            _, v_var = gp.var(self.diffs[np.where(b_ind != bins.size)])#account for lags bigger than bins
            return centers, n_bins, v, v_var
        else:
            return centers, n_bins, v


    def set_bins(self, bin_type, bins):
        """
        Calculate bin boundaries for bin parameters. See self.matheron for
        details.
        """
        if bin_type == "lin":
            db = (bins[1] - bins[0])/(bins[2] - 1)
            return np.linspace(bins[0],bins[1],int(bins[2])+1)-db/2
        elif bin_type == "bound":
            return np.asarray(bins)
        elif bin_type == "auto":
            return np.linspace(self.range[0],self.range[1]/2,int(bins)+1)

    def calc_lags(self):
        """
        Upon initialization, calculates distances between all points given in
        domain. Performed before any reductions are applied.
        """
        return pdist(self.x)

    def calc_diffs(self):
        """
        Upon initialization, calculates squared differences between all field
        values given. Performed before any reductions are applied.
        """
        c_indx = np.mask_indices(self.s, np.triu, k=1)
        diffs = (self.f[c_indx[0]] - self.f[c_indx[1]])**2
        return diffs

    def _c_reduce(f):
        #bin order, bin type etc
        @functools.wraps(f)
        def wrapper(self, typ, bnds, inplace):
            if bnds[0] > bnds [1]:
                raise Exception("Lower and upper bounds out of order.")
            if typ == "abs":
                if bnds[0] < self.range[0]:
                    warnings.warn("Lower bound smaller than smallest lag")
                if bnds[1] > self.range[1]:
                    warnings.warn("Upper bound greater than largest lag")
            elif typ == "quant":
                if not 0 < all(bnds) < 1:
                    raise Exception("Quantile bounds must be between 0 and 1")
            else:
                raise Exception("'{f}' not recognized bound type".fomat(f
                    = typ))

            return f(self, typ, bnds, inplace)
        return wrapper

    @_c_reduce
    def reduce(self, typ, bnds, inplace = True):#"abs", "quant"
        """
        *Reduce the lag domain of a given Variogram object. Removes data that
        lay outside the given lag boundaries. Very useful for reducing run
        times and memory requirements when performing further operations with
        variogram after initialization such as matheron. Object points and
        field values are not affected. Resulting Variogram objects are also
        marked with a self.reduced = True flag to indicate that the
        self.lags and self.diffs are not reflective of all combinations of
        self.x and self.f.

        Parameters
        ----------
        typ : str
            Descriptor of reduction criteria that is passed to the bnds
            parameter. Can be one of:
                * "abs" : eliminate data from self.diffs and self.lags that
                    come from lags greater or less than specified values
                * "quant" : eliminate data from self.diffs and self.lags by
                    quantiles.
        bnds : list
            2-element list to describe reduction criteria. Specific formats
            given below for each typ option.
                * "abs" : List in the format of (min, max) for range of lags
                    to be included after reduction.
                * "quant" : range of quantiles in format of (min, max) to keep
                    after reduction.
        inplace : bool
            Whether or not object is manipulated inplace. If False, will return
            variogram object with same x and f values but with the reduced lag
            domain.

        Returns
        -------
        new (optional) : Variogram
            New Variogram instance with same x and f values but reduced lag
            domain.
        """
        if typ  == "abs":
            min_lag = bnds[0]
            max_lag = bnds[1]
        elif trim_type == "quant":
            min_lag = np.quantile(self.lags, bnds[0])
            max_lag = np.quantile(self.lags, bnds[1])

        ids = np.where(min_lag <= self.lags <= max_lag)
        self.rm_ids(ids, inplace)


    def _c_rreduce(f):
        @functools.wraps(f)
        def wrapper(self, typ, amnt, inplace):
            if typ == "abs":
                if not 0 < amnt < self.lags.size:
                    raise Exception("'amnt' not between 0 and size of"
                                    " original array")
            elif typ == "frac":
                if not 0 < amnt < 1:
                    raise Exception("'amnt' must be between 0 and 1 for"
                                    " fraction reduction type")
            else:
                raise Exception("'{f}' not recognized bound type".format(f
                    = typ))
            return f(self, typ, amnt, inplace)
        return wrapper

    @_c_rreduce
    def rreduce(self, typ, amnt, inplace = False):
        """
        *Uniformly downsample stored lags and field value differences for
        faster calculations and smaller memory size. Object points and field
        values are not affected. Resulting Variogram objects are also marked
        with a self.reduced = True flag to indicate that the self.lags and
        self.diffs are not reflective of all combinations of self.x and self.f.

        Parameters
        ----------
        typ : str
            Descriptor of format of reduction amount passed to amnt. Can be
            one of:
                * "abs" : Size of remaining lag and field data difference
                array specified as an absolute size.
                * "frac" : Size of remaining lag and field data difference
                array specified as a fraction of original size.
        amnt : int, float
            Amount to reduce lag and field difference data vectors. Specific
            formats given below for each typ option.
                * "abs" : int giving the size of the remaining lag and
                    field data difference vectors.
                * "frac" : float between 0 and 1 describing how much of lag
                    and field difference should remain as fraction of original
                    size
        inplace : bool
            Whether or not object is manipulated inplace. If False, will return
            variogram object with same x and f values but with the reduced lag
            domain.

        Returns
        -------
        new (optional) : Variogram
            New Variogram instance with same x and f values but reduced lag
            domain.
        """
        if typ == "frac":
            size = int(amnt * self.lags.size)
        if typ == "abs":
            size = amnt

        ids = np.choice(self.lags.size, size)
        self.rm_ids(ids, inplace)



    def rm_ids(ids, inplace = False):
        """
        Helper function for reduction methods.
        """
        if inplace:
            self.lags = self.lags[ids]
            self.diffs = self.diffs[ids]
            self.range = (np.min(self.lags), np.max(self.lags))
            self.reduce = True
        else:
            new = Variogram(self.x, self.f)
            new.lags = new.lags[ids]
            new.diffs = new.diffs[ids]
            new.range = (np.min(new.lags), np.max(new.lags))
            new.reduce = True

            return new

    def check_init(self):
        """
        Notify user of errors during initialization of variogram
        """
        if self.x.shape[0] != self.f.shape[0]:
            raise Exception("Number of data points (x) and number of field "
                            "values (f) do not match.")
        if self.f.ndim > 1:
            raise Exception("Field values data (f) should be 1D")
        if self.x.ndim > 1:
            if self.x.shape[0] < self.x.shape[1]:
                warnings.warn("Dimension of domain exceeds number of data"
                    " points, check that each point is described by row")
        if np.any(np.iscomplex(self.x)):
            raise Exception("Complex numbers not taken in spatial data")
        if np.any(np.iscomplex(self.f)):
            raise Exception("Complex numbers not taken in field values")

    def cond_init(self):
        """
        Perform data format manimpulation where user interference not
        necessary.
        """
        self.x = np.asarray(self.x)
        self.f = np.asarray(self.f)
        if self.x.ndim < 2:
            self.x = self.x.reshape(self.x.size,1)




import numpy as np

def spherical(h, nug, sill, rang):
    variogram = nug + sill*np.piecewise(h, [h <= rang,  h > rang],[lambda h: 1.5*(h/rang)-0.5*(h/rang)**3,1])
    return variogram

def exponential(h, nug, sill, rang):
    variogram = nug +sill*(1.-np.exp(-3*h/rang))
    return variogram

def gaussian(h, nug, sill, rang):
    variogram =  nug+sill*(1.-np.exp(-(2*h/rang)**2))
    return variogram

mtags = {"sph" : spherical,
        "exp" : exponential,
        "gaus" : gaussian}
